// File_io
// enhanced storage uses fixed length fields, not all functions in this
// file are used
// ------------
// fio_init()
//
// For line orientated records, line ends with CR
// flOpen("filename") // creates if not exist
// flLines() // number of lines from 0
// flAdd("value")   // adds line on end
// flGet$(line) // gets line
//
// For inexed records
// fiNew(length,"filename") // create
// fUse("filename") // an existing file
// fiStore(index,"value")
// fiRead$(index)
// fClose() // any open file
//
// For 2 dimentional array, col size must be specified beforehand
// faNew(col,length,"filename")
// fUse("filename") // an existing file
// faPuts(row,col,"value")
// faGet(row,col)
// fClose() // any open file
constant FILEIO 1
//
constant DETECTSD 0
//constant SYS_FLAG 0x32c  // give > promt
constant SYS_COMBAUD 9600
constant NLENGTH 4 // length of a number
constant SMAX 1024 // maximum length of a string
constant FSTART 100 // past header record
constant TEMPIDX$ "_tmp.idn"
constant SDINS {PORTB(PORT),7}
constant EOT 0 // end of text marker
constant EOT$ "\000" // in octal
//
constant SYS_COMBAUD 9600
//
dim cf      // current file
//dim nrow,ncol,nsize    // 2d array maxroes and cols
// line otientated
dim xf  // current index file
dim ilines // keeps track number of lines
//
constant TTYPE 0
constant TROWS 1
constant TCOLS 2
constant TLEN  3
//  fileTop(0) = file type
//  fileTop(1) = row (not really used)
//  fileTop(2) = col
//  fileTop(3) = size
// Types:
//      Line  (fl) 	no head for this file
//      Index 3      (fi)
//      File array 4 (fa)
dim fileTop(11)

// =============================================================================
// =============================================================================
// *****************************************************************************
// returns 0 when card is inserted
// *****************************************************************************
function sd_check()
	if DETECTSD = 1 then
	    if io_pinGet(*SDINS()) = 1 then
	        print "\r\nNo SD Card"
	        return 1
	    else
	        return 0
	    endif
	else
		return 0
	endif
endf

// *****************************************************************************
// *****************************************************************************
function fio_init()
dim j
    cf = 0
    xf = 0
    ilines = 0
	for j = 0 to 9
		fileTop(j) = 0
 	next
    fileTop(3) = 80 // default
    io_pinRole(*SDINS(),IN,WPU)
    sd_check()
endf

// *****************************************************************************
// 1) right pads any string with spaces AFTER terminator
// 3) adds CR to end
// *****************************************************************************
function s_padright(*s$,nsize,delim)
dim sl=strlen(s$) // actual string length
dim j
	for j = sl+1 to nsize // after 0 terminator
		// +4 is string size
		pokec(?s$+4+j,delim)
	next
	pokec(?s$+4+nsize-1,10) //s$=s$+"\r"
endf

// =============================================================================
// UTILITY SAVES / GETS NUMBERS
// =============================================================================
// Numbers, integer and FP
// *****************************************************************************
// stores a integer number at the current file location, occupies 4 bytes
// *****************************************************************************
function f_putn(n,f)
dim a$, w
    poke(?a$+4,n)
    if fwrite(a$,NLENGTH,*w,f) <> 0 then
        print "Error writing number"
    endif
endf

// *****************************************************************************
// gets integer number from the given file location
// *****************************************************************************
function f_getn(f)
dim a$, w, rv=0
    if fread(a$,NLENGTH,*w,f) = 0 then
        rv = peek(?a$+4)
    else
        print "Error reading number"
    endif
    return rv
endf

// *****************************************************************************
// stores a fp number at the current file location, occupies 4 bytes
// *****************************************************************************
function f_putn#(n#,f)
dim a$, w
    poke(?a$+4,peek(?n#))
    if fwrite(a$,NLENGTH,*w,f) <> 0 then
        print "Error writing number"
    endif
endf

// *****************************************************************************
// gets fp number from the given file location
// *****************************************************************************
function f_getn#(f)
dim a$, w, rv#=0
    if fread(a$,NLENGTH,*w,f) = 0 then
        poke(?rv#,peek(?a$+4))
    else
        print "Error reading number"
    endif
    return rv#
endf

// *****************************************************************************
// strore a fixed length string at the current location, important to pass
// by ref as this will not destroy the padding
// *****************************************************************************
function f_puts(*s$,f)
dim w
    if fwrite(s$,fileTop(3),*w,f) <> 0 then
        print "Error writing string"
    endif
endf

// *****************************************************************************
// get a fixed length string at the current file location
// *****************************************************************************
function f_get$(f)
dim w, rv$[SMAX]
    if fread(rv$,fileTop(3),*w,f) <> 0 then
        print "Error reading string"
    endif
    return rv$
endf

// *****************************************************************************
// closes  curent file/s if open
// *****************************************************************************
function fClose()
    if cf <> 0 then
        fclose(cf)
    endif
    cf = 0
    if xf <> 0
        fclose(xf)
    endif
    xf = 0
endf


// =============================================================================
// file types, stored in head
//
// *****************************************************************************
// writes the array fileTop() to the start of the file as text so that when
// reading on another file system as text it will read okay
// *****************************************************************************
function  fs_puttop(fname$[80])
dim f, topLine$[128]="", j, w
	f = fopen(fname$,"r+")
	if f <> 0 then
		fseek(0,f) // make sure its at start
		for j = 0 to 9
			topLine$ = topLine$ + format$("%-9d,",fileTop(j))
		next
		fwrite(topLine$,FSTART,*w,f)
		fclose(f)
	else
		print format$("\r\nFile %s does not exist",fname$)
	endif
endf

// *****************************************************************************
// *****************************************************************************
function fs_gettop(fname$[80])
dim f, topLine$[128]="", j=0, p=0, w
	f = fopen(fname$,"r")
	if f <> 0 then
		fread(topLine$,FSTART,*w,f)
		while p >= 0
			fileTop(j) = str2int(token$(topLine$,*p,','))
			j = j + 1
		wend
		fclose(f)
	else
		print format$("\r\nFile %s does not exist",fname$)
	endif
endf


// *****************************************************************************
// creates a file for susiquent storage as 2 d array, used internally. Note
// that only the number of columns needs to be specified as the number of
// rows can grow as added
// returns 1 on okay
// *****************************************************************************
function fs_New(type,row,col,length,fname$[80])
dim rv = 0
    fClose() // close any open files
    cf = fopen(fname$,"w")
    if cf <> 0 then // okay
        fclose(cf)
        fileTop(0) = type // f_putn(type,cf)
        fileTop(1) = row // f_putn(row,cf)
        fileTop(2) = col //f_putn(col,cf)
        fileTop(3) = length // f_putn(length,cf)
		fs_puttop(fname$) // store to head of new file
//         cft = type
//         nrow = row
//         ncol = col
//         nsize = length
        cf = fopen(fname$,"r+") // ready to use
        rv = 1
    endif
    return rv
endf

// *****************************************************************************
// reads string at index - for internal use
// *****************************************************************************
function fi_get$(index)
dim rv$[SMAX]
    if sd_check() = 1 then
        return
    endif

    if cf = 0 then // not open
        print "\r\nfsUse first"
        return
    endif
    fseek((index*fileTop(3))+FSTART,cf)
    if fileTop(0) = 3 then // string okay
        rv$ =f_get$(cf)
    else
        print "\r\nFile wong type"
    endif
    return rv$
endf

// *****************************************************************************
// gets value
// This is always a string even though the storage may be different
// for 1d array set row to 1.
// *****************************************************************************
function fa_get$(row,col)
dim rv$[SMAX]
dim array_index
//    array_index = (ncol * nsize * row) + (col * nsize)
    array_index = (fileTop(2) * fileTop(3) * row) + (col * fileTop(3))
    if sd_check() = 1 then
        return
    endif

    if cf = 0 then // not open
        print "\r\nfsUse first"
        return
    endif

    fseek(array_index+FSTART,cf)
    // return astring
    if fileTop(0) = 4 then
        rv$ = f_get$(cf)
    else
        print "\r\nWrong file type"
    endif
    return rv$
endf


// =============================================================================
// =============================================================================
// =============================================================================
// USET SECTION
// =============================================================================
// *****************************************************************************
// file already created with fsNew() has details of type so how to write
// and read
// Applies to both indexed and arrayed fields
// *****************************************************************************
function fUse(fname$[80])
    if sd_check() = 1 then
        return
    endif

    fClose() // close any open files
    cf = fopen(fname$,"r+")
    if cf <> 0 then // does not exists
		fs_gettop(fname$) // gets file details
    else
        print "\r\nCreate file first"
    endif
endf

// *****************************************************************************
// returns number of rows for the current file
// *****************************************************************************
function fRows()
dim rv=0
	if fileTop(TTYPE) <> 0 then
		if fileTop(TTYPE) = 3 then // index type
			rv = (fsize(cf) - FSTART) / fileTop(TLEN)
		endif
		if fileTop(TTYPE) = 4 then // array
			rv = (fsize(cf) - FSTART) / (fileTop(TCOLS) * fileTop(TLEN))
		endif
	endif
	return rv
endf

// *****************************************************************************
// converts file to 'char' spearated format, smaller because it removes the
// padding
// *****************************************************************************
function fConvert(infile$[40],outfile$[40])
dim rows, cols, a$[SMAX], fo, w, n
	fUse(infile$) // the file
	if fileTop(TTYPE) <> 0 then // 0 is line and cant format that
		fo = fopen(outfile$,"w")
		if fileTop(TTYPE) = 3 then // index type
			n = fRows()
			for rows = 0 to n-1 // in current file
				a$ = fi_get$(rows)+"\r\n"
				fwrite(a$,strlen(a$),*w,fo)
			next
			// last one without delimiter
			a$ = fi_get$(rows)
			fwrite(a$,strlen(a$),*w,fo)
			fclose(fo)
		endif
		if fileTop(TTYPE) = 4 then // 2d type
			for rows = 0 to fRows()-1 // in current file
				for cols = 0 to fileTop(TCOLS)-2 // less last one and from 0
					a$ = fa_get$(rows,cols)+","
					fwrite(a$,strlen(a$),*w,fo)
				next
				// last one without delimiter and add CRLF
				a$ = fa_get$(rows,cols+1)+"\r\n"
                fwrite(a$,strlen(a$),*w,fo)
			next
			fclose(fo)
		endif
	endif
endf

// =============================================================================
// =============================================================================
// INDEXED FILES
// =============================================================================
// *****************************************************************************
// defult file is string, user to specify max length of string, this kind
// of file is only limited by the size of the sd card - no need to specify
// max columns or rows
// *****************************************************************************
function fiNew(length,fname$[80])
    if sd_check() = 1 then
        return
    endif

    if fs_New(3,0,1,length,fname$) = 0 then
        print "\r\nError creating",fname$
    endif
endf

// *****************************************************************************
// stores value to given index
// *****************************************************************************
function fiPut(index,value$[SMAX])
    if sd_check() = 1 then
        return
    endif

    if cf = 0 then // not open
        print "\r\nfsUse first"
        return
    endif
    fseek((index*fileTop(3))+FSTART,cf)
    if fileTop(0) = 3 then // string okay
		s_padright(*value$,fileTop(TLEN),' ')
        f_puts(*value$,cf)
    else
        print "\r\nFile wong type"
    endif
    fflush(cf) // write
endf

// *****************************************************************************
// reads string at index
// *****************************************************************************
function fiGet(index)
    print fi_get$(index)
endf


// =============================================================================
// =============================================================================
// LINE ORIENTED FILES
// =============================================================================
// creates a temporary index file so movement through lines is quicker

// *****************************************************************************
// temporary index file is a list of integer numbers equating line numbers
// with file positions. This is so that the line number can be quickley found
// *****************************************************************************
function create_temp_index(xfp,ifp)
dim a$
    f_putn(0,xfp) // first line always 0
    ilines = 0
    // go through file to index
    while feof(ifp) = 0
        fgets(a$,ifp)
        //if asc(a$,0) < 14 then // ignore blank lines
            f_putn(fpos(ifp),xfp)
            ilines = ilines + 1
        //endif
    wend
    fflush(xfp)
endf

// *****************************************************************************
// using a previously created indexfile, gets the fpos or a particular
// line number. Uses previously opened idex file xf
// *****************************************************************************
function getLineStart(ln)
dim  rv
    fseek(ln*4,xf)
    rv = f_getn(xf)
    print rv
endf

// *****************************************************************************
// *****************************************************************************
function flNew(fname$[80])
    if sd_check() = 1 then
        return
    endif

    fClose() // close any open files
    cf = fopen(fname$,"w")
    fclose(cf)
    cf = fopen(fname$,"r+") // re-open for read and write
    xf = fopen(TEMPIDX$,"w") // create indexfile
    create_temp_index(xf,cf)
    fclose(xf)
    xf = fopen(TEMPIDX$,"r+") //  re-open for append
endf

// opens an existing file, if no file exists then creates one.
// *****************************************************************************
function flUse(fname$[80])
    if sd_check() = 1 then
        return
    endif

    fClose() // close any open files
    cf = fopen(fname$,"r+")
    if cf = 0 then
        print "\r\nFile does not Exist"
	else
	    xf = fopen(TEMPIDX$,"w") // create indexfile
    	create_temp_index(xf,cf)
    	fclose(xf)
    	xf = fopen(TEMPIDX$,"r+") //  re-open for append
	endif
endf

// *****************************************************************************
// returns number of lines, includes 0 so last line is fllines - 1
// *****************************************************************************
function flLines()
    if sd_check() = 1 then
        return
    endif

    print ilines
endf

// *****************************************************************************
// gets requested line
// *****************************************************************************
function flGet(line)
dim rv$[SMAX]
    if sd_check() = 1 then
        return
    endif

    fseek(getLineStart(line),cf)
    fgets(rv$,cf)
    print rv$
endf

// *****************************************************************************
// adds a line on to the end of the file
// *****************************************************************************
function flAdd(value$[SMAX])
dim w
    if sd_check() = 1 then
        return
    endif

    fseek(fsize(cf),cf) // move to end of file
    // update index
    f_putn(fpos(cf),xf)
    fflush(xf)
    // add value
    s_padright(*value$,fileTop(TLEN),' ') // record size

    if fwrite(value$,fileTop(3),*w,cf) <> 0 then
        print "\r\nError writing ",value$
    else
        fflush(cf)
        if strlen(value$) > 0 then
            ilines = ilines + 1
        endif
    endif
endf



// =============================================================================
// =============================================================================
// TWO DIMENTIONAL ARRAY STORAGE
// =============================================================================
// *****************************************************************************
// creates a 2d array, NOTE the file size only grows when more rows are added
// the initial size is very small
// no need to specify number of rows
// *****************************************************************************
function faNew(col,length,fname$[80])
    if sd_check() = 1 then
        return
    endif

    if fs_New(4,1,col,length,fname$) = 0 then
        print "\r\nError creating",fname$
    endif
endf

// *****************************************************************************
// puts string to file in requested position as 2d array
// array_index = (ncol * size * row) + (col * size)
// *****************************************************************************
function faPut(row,col,value$[SMAX])
dim array_index
    if sd_check() = 1 then
        return
    endif

    if cf = 0 then // not open
        print "\r\nfsUse first"
        return
    endif
    if col > fileTop(2) then
        print "\r\nCol value too large"
        return
    endif
//    array_index = (ncol * nsize * row) + (col * nsize)
	array_index = (fileTop(2) * fileTop(3) * row) + (col * fileTop(3))

//      print "\r\nnrow=",fileTop(1)
//      print "\r\nncol=",fileTop(2)
//      print "\r\nnsize=",fileTop(3)
//      print "\r\narray_index=",array_index
//      print "\r\nFSTART=",FSTART

    fseek(array_index+FSTART,cf) // move to field
    // how to store depends on type
    if fileTop(0) = 4 then // string
		s_padright(*value$,fileTop(TLEN),' ') // pad with spaces + CR on end
        f_puts(*value$,cf)
    else
        print "\r\nFile wong type"
    endif
    fflush(cf)
endf

// *****************************************************************************
// gets value
// This is always a string even though the storage may be different
// for 1d array set row to 1.
// *****************************************************************************
function faGet(row,col)
    print fa_get$(row,col)
endf


// *****************************************************************************
// *****************************************************************************
function main()
    fio_init()
endf

