// Rookie 3
// Contains:
// Interrupts
// Ports
// ADC does not use auto convert and has timeout
// Timers
// improvements:
// 1) handles timers better
// 2) unifies ports and interrupts
// 3) adc now takes into account 44 pin devices
// 4) less duplication due to use of constant arrays
// 5) just change constant arrays for device change
// Timers fully tested
// Interrupt on change not tested
// ports not tested
// 29 Dec 2015
//    added spi_init16(speed)
//    added spi_init32(speed)
constant ROOKIE3_1 1
// ======================= constants determine device ==========================
constant CLR    4
constant SET    8
constant OUT    4
constant IN     8
constant WOFF   0 // waek pull up / down off
constant WPU    1 // weak pull up
constant WPD    2 // weak pull down
//
// =========================== ADC =============================================
// The 28 pin device has ADC channels as follows
// On Port A
// AN0  RA0
// AN1  RA1
// AN2 RB0
// AN3 RB1
// AN4 RB2
// AN5 RB3
// AN6 RC0   ** 44 pin
// AN7 RC1   ** 44 pin
// AN8 RC2   ** 44 pin
// AN9  RB15
// AN10 RB14
// AN11 RB13
// AN12 RB12
// This is a total of 10 channels 0 to 9
// NOTE that channels 6,7, and 8 do not exist
//constant SAMPLE 5
//constant CONVERSION 5
// the high byte is the port, 0=A, 1=B, 2=C, 3=error
// low byte is the port pin, table index is channel number
constant CHANMAP {0x00,0x01,0x10,0x11,0x12,0x13,0x20,0x21,0x22,0x30, \
0x30,0x30,0x1f,0x1e,0x1d,0x1c }
//
constant AD1CON1         0xBF809000
constant AD1CON2         0xBF809010
constant AD1CON3         0xBF809020
constant AD1CHS          0xBF809040
constant AD1CSSL         0xBF809050
constant ADC1BUF0        0xBF809070
//
// =============== Timer compound information ============================
// timer parameters {ID,IECn,IFSn,Iflag,IPCn,TnCON,TMRn,PRn}
// ID 0x1 to 0xf
constant RID        0
constant IREN       1 // e.g. IEC0
constant IRFLAG     2 // e.g. IFS0
constant IRMASK     3 // e.g 1 << 14
constant IRPRI      4 // priority register PRI
constant TCON       5
constant TMR        6
constant PR         7
constant TIMER1 {1,0xBF881060,0xBF881030,1<<4,0xBF8810A0,0xBF800600, \
0xBF800610,0xBF800620}
constant TIMER2 {2,0xBF881060,0xBF881030,1<<9,0xBF8810B0,0xBF800800, \
0xBF800810,0xBF800820}
constant TIMER3 {3,0xBF881060,0xBF881030,1<<14,0xBF8810C0,0xBF800A00, \
0xBF800A10,0xBF800A20}
constant TIMER4 {4,0xBF881060,0xBF881030,1<<19,0xBF8810D0,0xBF800C00, \
0xBF800C10,0xBF800C20}
constant TIMER5 {5,0xBF881060,0xBF881030,1<<24,0xBF8810E0,0xBF800E00, \
0xBF800E10,0xBF800E20}
constant TIMER23 {6,0xBF881060,0xBF881030,1<<14,0xBF8810C0,0xBF800800, \
0xBF800810,0xBF800820}
constant TIMER45 {7,0xBF881060,0xBF881030,1<<24,0xBF8810E0,0xBF800C00, \
0xBF800C10,0xBF800C20}
//
// ============================ PORTS ==========================================
// PORT parameters  {ID,IECn,IFSn,Iflag,IPCn}
// ID   0x10 porta
//      0x20 portb
//      etc
// relative offsets from ANSEL
constant PORT           5   // location of port(ansel) in param list
constant ADC_OFF        0
constant TRIS_OFF       0x10
constant PORT_OFF       0x20
constant LAT_OFF        0x30
constant PUP_OFF        0x50 // weak pull up
constant PDOWN_OFF      0x60 // weak pull down
constant CNC_OFF        0x70 // change notification control
constant CNE_OFF        0x80 // chnage notification enable
// PORTA {ID,IEC1,IFS1,IPC8,flag,ANSELA} // change notice flag
// PORTB {ID,IEC1,IFS1,IPC8,falg,ANSELA}
constant PORTA {10,0xBF881070,0xBF881040,1<<13,0xBF881110,0xBF886000}
constant PORTB {20,0xBF881070,0xBF881040,1<<14,0xBF881110,0xBF886100}
constant PORTC {30,0xBF881070,0xBF881040,1<<15,0xBF881110,0xBF886200}
//
// Port pins - for 44 pin QFP
constant PPORT  0
constant PVAL   1
constant RA0    {PORTA(PORT),0}
constant RA1    {PORTA(PORT),1}
constant RA2    {PORTA(PORT),2}
constant RA3    {PORTA(PORT),3}
constant RA4    {PORTA(PORT),4}
constant RA7    {PORTA(PORT),7}
constant RA8    {PORTA(PORT),8}
constant RA9    {PORTA(PORT),9}
constant RA10   {PORTA(PORT),10}
constant RB0    {PORTB(PORT),0}
constant RB1    {PORTB(PORT),1}
constant RB2    {PORTB(PORT),2}
constant RB3    {PORTB(PORT),3}
constant RB4    {PORTB(PORT),4}
constant RB5    {PORTB(PORT),5}
constant RB6    {PORTB(PORT),6}
constant RB7    {PORTB(PORT),7}
constant RB8    {PORTB(PORT),8}
constant RB9    {PORTB(PORT),9}
constant RB10   {PORTB(PORT),10}
constant RB11   {PORTB(PORT),11}
constant RB12   {PORTB(PORT),12}
constant RB13   {PORTB(PORT),13}
constant RB14   {PORTB(PORT),14}
constant RB15   {PORTB(PORT),15}
constant RC0    {PORTC(PORT),0}
constant RC1    {PORTC(PORT),1}
constant RC2    {PORTC(PORT),2}
constant RC3    {PORTC(PORT),3}
constant RC4    {PORTC(PORT),4}
constant RC5    {PORTC(PORT),5}
constant RC6    {PORTC(PORT),6}
constant RC7    {PORTC(PORT),7}
constant RC8    {PORTC(PORT),8}
constant RC9    {PORTC(PORT),9}
//
// ========================== INTERRUPTS =======================================
// Note interrupt infomation is within ports and timers
// for clearing interrupts
constant IEC0            0xBF881060
constant IEC1            0xBF881070

// ==================== END of CONSTANTS =======================================

// plugins for interrupt global enable / disable
constant di_pi { 0x27bdfff0, 0xafbe000c, 0xafb00008, 0x03a0f021, \
0xafc40010, 0xafc00000, 0x41706000, 0xafd00000, 0x8fc20000, 0x03c0e821, \
0x8fbe000c, 0x8fb00008, 0x03e00008, 0x27bd0010 }

constant ei_pi { 0x27bdfff0, 0xafbe000c, 0xafb00008, 0x03a0f021, \
0xafc40010, 0xafc00000, 0x41706020, 0xafd00000, 0x8fc20000, 0x03c0e821, \
0x8fbe000c, 0x8fb00008, 0x03e00008, 0x27bd0010 }

// =============================================================================
// =============================================================================
// INTERRUPTS
//
// *****************************************************************************
// set intrrupt example usage
// ir_set(*TIMER2(),3,"irq")
// *****************************************************************************
function ir_set(*ir(),pri,func$[40])
    // check not already enabled
    if (peek(ir(IREN)) & ir(IRMASK)) = 0 then
        poke(ir(IRFLAG)+CLR,ir(IRMASK)) // clear interrupt
        poke(ir(IRPRI)+SET,pri << 2) // set priority
        poke(ir(IREN)+SET,ir(IRMASK)) // enable
        irset(ir(IRFLAG),ir(IRMASK),func$) // table entry
    endif
endf

// *****************************************************************************
// Clears the interrupt table and also the flag used in rookie so that ir_set
// can be called more than once without causing problems
// *****************************************************************************
function ir_clear()
    poke(IEC0,0)
    poke(IEC1,0)
    irclear()
endf

// *****************************************************************************
// globaly enables and disables interrupts
// *****************************************************************************
function ir_di()
    call(?di_pi(0),0)
endf

function ir_ei()
    call(?ei_pi(0),0)
endf

// =============================================================================
// =============================================================================
// PORTS
//
//
// *****************************************************************************
// sets i/o pin to either input or output and also sets the weak pull up or
// pull down
// **** uses actualpin *********
// Example: set RB5 to be an output
// io_pinMode(*RB5(),OUT,POFF)
// *****************************************************************************
function io_pinRole(*pin(),ionorout,pull)
      poke(pin(PPORT)+ADC_OFF+CLR,1 << pin(PVAL)) // set adc
      poke(pin(PPORT)+TRIS_OFF+ionorout,1 << pin(PVAL)) // set TRISn
      // handle weak pull up/down
      poke(pin(PPORT)+PUP_OFF+CLR, 1 << pin(PVAL)) // both off
      poke(pin(PPORT)+PDOWN_OFF+CLR, 1 << pin(PVAL))
      if pull = WPU then
          poke(pin(PPORT)+PUP_OFF+SET, 1 << pin(PVAL))
      endif
      if pull = WPD then
          poke(pin(PPORT)+PDOWN_OFF+SET, 1 << pin(PVAL))
      endif
endf
//
// *****************************************************************************
// sets pin to LOW,HIGH or TOGGLE (0,1,2)
// *****************************************************************************
function io_pinSet(*pin(),value)
dim mode = (value * 4) + 4 // make CLR,SET or toggle
    poke(pin(PPORT)+LAT_OFF+mode,1 << pin(PVAL))
endf
//
// *****************************************************************************
// gets pin value
// *****************************************************************************
function io_pinGet(*pin())
dim k
    k = peek(pin(PPORT)+PORT_OFF) // PORTn
    return (k >> pin(PVAL) & 0x1
endf
//
// *****************************************************************************
// gets a whole port into a 32 bit integer
// Example:
// x = io_getPort(*PORTA())
// *****************************************************************************
function io_getPort(*port())
    return peek(port(PORT)+PORT_OFF)
endf

// *****************************************************************************
// sets a whole port to a value
// Example set bits 0 and 1 high and the rest low of port B
// io_setPort(*PORTB(),3)
// *****************************************************************************
function io_setPort(*port(),value)
    poke(port(PORT)+LAT_OFF,value)
endf
//
// *****************************************************************************
// sets or clears only the bits that are set on the value
// *****************************************************************************
function io_setPortH(*port(),value)
    poke(port(PORT)+LAT_OFF+SET,value)
endf
//
function io_setPortL(*port(),value)
    poke(port(PORT)+LAT_OFF+CLR,value)
endf
//
// *****************************************************************************
// sets intruupt on change for a pin
// *****************************************************************************
function ir_io(*ir(), pri, func$[40], pin)
    ir_set(*ir(0),pri, func$)
    if (ir(RID) & 0xf0) <> 0 then
        // port change notice
        poke(ir(PORT)+CNE_OFF+SET, 1 << pin)
        poke(ir(PORT)+CNC_OFF+SET, 1 << 15)  // turn on
    endif
endf

// =============================================================================
// =============================================================================
// ADC
//
constant    ADCSAMP 2
constant    ADCDONE 1
// *****************************************************************************
// Set up ADC for chanel, does not use auto conversion
// This is called once
// *****************************************************************************
function adc_init(chan)
dim map, ch
	// map channel to actual port high byte 0=1,1=b
	// low byte is port
	map = CHANMAP(chan)
	if map > 0x2f then
	   print "\nERROR chanel does not exist"
	   return
	endif
	ch = map & 0xf
	select(map & 0xf0)
		case(0x00)
    		poke(PORTA(PORT)+TRIS_OFF+SET,1 << ch) // set as i/p
    		poke(PORTA(PORT)+SET,1 << ch) // set as analogue
    		break
    	case(0x10)
    		poke(PORTB(PORT)+TRIS_OFF+SET,1 << ch) // set as i/p
    		poke(PORTB(PORT)+SET,1 << ch) // set as analogue
    		break
    	case(0x20)
    		poke(PORTC(PORT)+TRIS_OFF+SET,1 << ch) // set as i/p
    		poke(PORTC(PORT)+SET,1 << ch) // set as analogue
    		break
    	default
    		print "\nError with channel number"
    endselect
    @AD1CON3 = 0 // reset first, clock derived from PBCLK
    @AD1CON3 = 0 // (SAMPLE << 8) | CONVERSION
    @AD1CON2 = 0
    @AD1CON1 = 0x8000 // 0x80e0 // auto conversion + turn on ADC
    @AD1CSSL = 0
endf
//
// *****************************************************************************
// More control over adc, sample is the sampling period and timeout will
// return if something is wrong with the conversion
// times are arbitory 1,000,000 gives about 1 second
// example: adc_get(1,100,10000)
// *****************************************************************************
function adc_get(chan,sample,timeout)
dim t,j
    @AD1CHS = chan << 16   // activate requested ADC pin
    poke(AD1CON1+SET, ADCSAMP) // start ADC sampling
    for j = 0 to sample; next // sample period
    poke(AD1CON1+CLR, ADCSAMP) // stop sample and start conversion
    // wait for converstion to end with timeout
    while (peek(AD1CON1) & ADCDONE) = 0
        timeout = timeout - 1
        if timeout <= 0 then
            return 0
        endif
    wend
	poke(AD1CON1+CLR, ADCDONE) // clear flag
	t = @ADC1BUF0 // get value
    return t
endf
// =============================================================================
// =============================================================================
// TIMERS
//
// Calculater period assuming 40MHz perif clock
// prescale             counts / ms     second   *32 bits needed  Timer 1 pre
// 256 (7) =  6.4 us    156.25          156250                        3
// 64  (6) =  1.6 us    625             625000   *                    2
// 32  (5) =  0.8 us    1250            1250000  *
// 16  (4) =  0.4 us    2500            2500000  *
// 8   (3) =  0.2 us    5000            5000000  *                    1
// 4   (2) =  0.1 us    10000           10000000 *
// 2   (1) =  0.05 us   20000           20000000 *
// 1   (0) =  0.025us   40000           40000000 *                    0
//
// *****************************************************************************
// tmr is an array constant
// prescale is a number form 0 to 7 (0 to 3 for timer 1)
// limit is the count up register before rollover
// example of use
// tmr_init(*TIMER2(),7,625)
// *****************************************************************************
function tmr_init(*tmr(),prescale,limit)
dim t=0
    if (tmr(RID) = 1) && (prescale > 3) then
        print "\nError, max prescale is 3, setting to 3"
        prescale=3
    endif

    // check for double timers
    if (tmr(RID) = 6) || (tmr(RID) = 7) then
        t = 8
    endif

    t = t + 0x8000 | (prescale << 4)
    poke(tmr(TCON),t)
    poke(tmr(TMR),0)
    poke(tmr(PR),limit)
endf

// *****************************************************************************
// utilitiies, example of use
// tmr_stop(*TIMER2())  // stops timer 2
// *****************************************************************************
function tmr_stop(*tmr())
    poke(tmr(TCON)+CLR,0x8000)
endf
function tmr_start(*tmr())
    poke(tmr(TCON)+SET,0x8000)
endf
function tmr_get(*tmr())
    return peek(tmr(TMR))
endf
function tmr_clr(*tmr())
    poke(tmr(TMR),0)
endf
function tmr_limit(*tmr(),limit)
    poke(tmr(PR),limit)
endf

// =============================================================================
// I2C Section
// Hardware
// SCL1 is RB8
// SDA1 is RB9
// SDA2 is RB2
// SCL2 is RB3
//
// MX1 I2C regitsers used
constant I2C1CON         0xBF805000
constant I2C1STAT        0xBF805010
constant I2C1BRG         0xBF805040
constant I2C1TRN         0xBF805050
constant I2C1RCV         0xBF805060
constant CCLK    40000000
constant I2CTIMEOUT 100000
// Config
constant SCL1     1 << 8
constant SDA1    1 << 9
constant SEN    1 << 0
constant RSEN     1 << 1
constant PEN    1 << 2
constant RCEN    1 << 3
constant ACKEN    1 << 4
constant ACKDT    1 << 5
constant CONSTAT SEN | RSEN | PEN | RCEN | ACKEN
// status
constant I2COV        1 << 6
constant IWCOL         1 << 7
constant ACKSTAT     1 << 15
constant TRSTAT     1 << 14
constant RBF        1 << 1
constant TBF         1 << 0
//
// ============================= I2C Channel 1 =================================
// The Functions here are set up for channel 1. To use channel 2 change all
// of the I2C1xxx for I2C2xxx and use the channel 2 pins 7 and 8.
//
// General use:
// i2c_start(address) // even address for writing, odd for reading
// i2c_putc(data)
// i2c_stop()

// *****************************************************************************
// waits for interface to complete a command
// *****************************************************************************
function i2c_idle()
dim tout = I2CTIMEOUT
    while (peek(I2C1CON) & CONSTAT) <> 0
       tout = tout -1
       if tout < 1 then
           break
       endif
    wend
    tout = I2CTIMEOUT
    while (peek(I2C1STAT) & TRSTAT) <> 0
       tout = tout -1
       if tout < 1 then
           break
       endif
    wend
endf

// *****************************************************************************
// Open i2c, specify rate in Hz
// *****************************************************************************
function i2c_open(rate)
dim tmp=SCL1 | SDA1
    @I2C1BRG = (CCLK/2/rate)-2
    poke(PORTB(PORT)+TRIS_OFF+SET,tmp) // set pins to i/p
    @I2C1CON = 0x8000  // on
endf

// *****************************************************************************
// Close
// *****************************************************************************
function i2c_close()
    @I2C1CON = 0x0  // off
endf

// *****************************************************************************
// send re-start condition
// *****************************************************************************
function i2c_rstart()
    poke(I2C1CON+SET,RSEN)
    i2c_idle()
endf

// *****************************************************************************
// send stop condition
// *****************************************************************************
function i2c_stop()
    poke(I2C1CON+SET,PEN)
    i2c_idle()
endf

// *****************************************************************************
// searches bus for I2C device, returns when address is found, returns with the
// found address.
// specify start and end address
// *****************************************************************************
function i2c_find(start,end)
dim j, rv=0
    i2c_open(100000)
    for j = start to end
        poke(I2C1CON+SET,SEN) // start
        i2c_idle()
        @I2C1TRN=j
        wait(1)
        if peek(I2C1STAT) = 8 then
            rv = j
            break
        endif
        poke(I2C1CON+CLR,0x8000)
        poke(I2C1CON+SET,0x8000)
    next
    poke(I2C1CON+CLR,0x8000)
    poke(I2C1CON+SET,0x8000)
    return rv
endf

// *****************************************************************************
// send byte to i2c
// returns     -1 on error
//            1 if NACK recieved (normal)
//            0 if ACK received
// *****************************************************************************
function i2c_putc(c)
dim a
    @I2C1TRN = c
    if (@I2C1STAT & IWCOL) <> 0 then
        return -1
    endif
    // wait for buffer to empty
    while (@I2C1STAT & TBF) <> 0; wend
    i2c_idle()
    if @I2C1STAT & ACKSTAT = 0 then
        return 0
    else
        return 1
    endif
endf

// *****************************************************************************
// send start condition and address
// *****************************************************************************
function i2c_start(adr)
    poke(I2C1CON+SET,SEN)
    i2c_idle()
    i2c_putc(adr)
endf

// *****************************************************************************
// recieve byte from I2C, error indiocation by print
// *****************************************************************************
function i2c_getc(last)
dim tout = I2CTIMEOUT, c
    poke(I2C1CON+SET,RCEN)
    while (peek(I2C1STAT) & RBF) <> RBF
        tout = tout - 1
        if tout < 1 then
        	print "Error on I2C getc"
            return 0
        endif
    wend
    c = @I2C1RCV  // received data
    // send ack depending on last
    if last = 1 then
        // nack
        poke(I2C1CON+SET,ACKDT)
        poke(I2C1CON+SET,ACKEN)
    else
        // ack
        poke(I2C1CON+CLR,ACKDT)
        poke(I2C1CON+SET,ACKEN)
    endif
    // wait until ACK/NAC is over
    while (peek(I2C1CON) & ACKEN) <> 0; wend
    // check returned
    if @I2C1STAT & 0x8000 = 0x8000 then
        return c
    else
    	print "Error on I2C getc"
        return 0
    endif
endf

// =============================================================================
// SPI Section
//
// Hardware for SPI
// Data out SDO1 is RB13
// Data in  SDI1 is RB8
// Clock SCK1 is RB14
//
constant SPI1CON         0xBF805800
constant SPI1STAT        0xBF805810
constant SPI1BUF         0xBF805820
constant SPI1BRG         0xBF805830
constant SPI1CON2        0xBF805840
// PPS
constant SDI1R           0xBF80FA84
constant RPB13R          0xBF80FB60
constant PPSSDI 4 // this is RB8 on group 2
constant PPSSD0 3 // group 3
constant SPI1RBUF 1 << 11 // set when spi is busy

// *****************************************************************************
// initialise PSI hardware, set speed in Hz
// MAX is 20MHz
// Set clock polarity etc. to suit, this is set in SPI1CON
// *****************************************************************************
function spi_init(speed)
dim sp
    // spi interface
    sp = (40000000 / (2*speed)) - 1
    if sp <= 0 then
        sp = 1
    endif
    if sp > 1024 then
        sp = 1024
    endif
    @SPI1STAT = 0
    @SPI1CON = 0x8120    // enable | CKE=1 | MASTER (8 bit mode)
    @SPI1CON2 = 0
    @SPI1BRG = sp
    // PPS SPI pins are mapped
    @SDI1R = PPSSDI
    @RPB13R = PPSSD0
endf

// *****************************************************************************
// 16 bit spi
// *****************************************************************************
function spi_init16(speed)
    spi_init(speed)
    @SPI1CON = 0x8520    // enable | CKE=1 | MASTER (16 bit mode)
endf

// *****************************************************************************
// 32 bit spi
// *****************************************************************************
function spi_init32(speed)
    spi_init(speed)
    @SPI1CON = 0x8920    // enable | CKE=1 | MASTER (32 bit mode)
endf

// *****************************************************************************
// main data transfer call to read or write, when reading set v to -1
// *****************************************************************************
function spi_data(v)
    @SPI1BUF = v
    while (peek(SPI1STAT) & SPI1RBUF) <> 0;wend
    return @SPI1BUF
endf
