// rookie.bas PART 2
// Load this and save to flash. It is intended for first time users of ByPic
// that makes it much simpler for basic tasks to be carried out
// Processor BV500 - MX1
// Version see main()
// June 2013 * error corrected in pps_out(pin, periph)
//
//
constant ROOKIE3_2 1

// =============================================================================
// PPS Peripheral Pin Select
//
// The following constants belong to a group 0 to 3 which is checked by the
// required peripheral pin when setting a PPS
// Format, first value high byte is group, second value is select. The second
// value is the port pin register used for output, i.e. RPA0R
// group 0
constant    PRA0    {0x00, 0xBF80FB00}
constant    PRB3    {0x01, 0xBF80FB38}
constant    PRB4    {0x02, 0xBF80FB3C}
constant    PRB15   {0x03, 0xBF80FB68}
constant    PRB7    {0x04, 0xBF80FB48}
constant    PRC7    {0x05, 0xBF80FB88}
constant    PRC0    {0x06, 0xBF80FB6C}
constant    PRC5    {0x07, 0xBF80FB80}
// group 1
constant    PRA1    {0x10, 0xBF80FB04}
constant    PRB5    {0x11, 0xBF80FB40}
constant    PRB1    {0x12, 0xBF80FB30}
constant    PRB11   {0x13, 0xBF80FB58}
constant    PRB8    {0x14, 0xBF80FB4C}
constant    PRA8    {0x15, 0xBF80FB20}
constant    PRC8    {0x16, 0xBF80FB8C}
constant    PRA9    {0x17, 0xBF80FB24}
// group 2
constant    PRA2    {0x20, 0xBF80FB08}
constant    PRB6    {0x21, 0xBF80FB44}
constant    PRA4    {0x22, 0xBF80FB10}
constant    PRB13   {0x23, 0xBF80FB60}
constant    PRB2    {0x24, 0xBF80FB34}
constant    PRC6    {0x25, 0xBF80FB84}
constant    PRC1    {0x26, 0xBF80FB70}
constant    PRC3    {0x27, 0xBF80FB78}
// group 3
constant    PRA3    {0x30, 0xBF80FB0C}
constant    PRB14   {0x31, 0xBF80FB64}
constant    PRB0    {0x32, 0xBF80FB2C}
constant    PRB10   {0x33, 0xBF80FB54}
constant    PRB9    {0x34, 0xBF80FB50}
constant    PRC9    {0x35, 0xBF80FB90}
constant    PRC2    {0x36, 0xBF80FB74}
constant    PRC4    {0x37, 0xBF80FB7C}
// The folloing are the peripheral constants, these are tables as they need to
// be checked for the correct group, format: <group><register>
// INPUT
// group 0
constant    INT4    {0x00, 0xBF80FA10}
constant    T2CK    {0x00, 0xBF80FA18}
constant    IC4     {0x00, 0xBF80FA34}
constant    SS1     {0x00, 0xBF80FA88}
constant    REFCLKI {0x00, 0xBF80FAB8}
// group 1
constant    INT3    {0x10, 0xBF80FA0C}
constant    T3CK    {0x10, 0xBF80FA1C}
constant    IC3     {0x10, 0xBF80FA30}
constant    U1CTS   {0x10, 0xBF80FA54}
constant    U2RX    {0x10, 0xBF80FA58}
constant    SDI1    {0x10, 0xBF80FA84}
// group 2
constant    INT2    {0x20, 0xBF80FA08}
constant    T4CK    {0x20, 0xBF80FA20}
constant    IC1     {0x20, 0xBF80FA28}
constant    IC5     {0x20, 0xBF80FA38}
constant    U1RX    {0x20, 0xBF80FA50}
constant    U2CTS   {0x20, 0xBF80FA5C}
constant    SDI2    {0x20, 0xBF80FA90}
constant    OCFB    {0x20, 0xBF80FA4C}
// group 3
constant    INT1    {0x30, 0xBF80FA04}
constant    T5CK    {0x30, 0xBF80FA24}
constant    IC2     {0x30, 0xBF80FA2C}
constant    SS2     {0x30, 0xBF80FA94}
constant    OCFA    {0x30, 0xBF80FA48}
// OUTPUTS
// the format for this is that the position in the table represents the group
// and the number is the select, $ is used to save space and it also makes it
// different from the above groups as some names repeat
constant    U1TXs   {1,0,0,1}
constant    U2RTSs  {2,0,0,1}
constant    SS1s    {3,0,0,0}
constant    OC1s    {5,0,0,0}
constant    C2OUTs  {7,0,0,0}
constant    SD01s   {0,3,3,0}
constant    SD02s   {0,4,4,0}
constant    OC2s    {0,5,0,0}
constant    OC4s    {0,0,5,0}
constant    OC5s    {0,0,6,0}
constant    REFCLK0s {0,0,7,0}
constant    U2TXs   {0,0,0,2}
constant    PSS2s    {0,0,0,4}
constant    OC3s    {0,0,0,5}
constant    C1OUTs  {0,0,0,7}
constant    NOCONNECTs {10}
//
// *****************************************************************************
// Peripheral pin select input
// example pps_in(*INT4(), *PRB4())
// *****************************************************************************
function pps_in(*periph(), *pin())
    // check for correct group
    if periph(0) = (pin(0) & 0xf0) then
        poke(periph(1),(pin(0) & 0xf))
    else
        print "\nGroup mismatch"
    endif
endf

// *****************************************************************************
// Peripheral pin select output
// example pps_in(*PRB4(), *U1TXs())
// *****************************************************************************
function pps_out(*pin(), *periph())
dim group
    if periph(0) = 10 then // special case disconnect
        poke(pin(1),0)
    else
        // check for correct group
        group = pin(0) >> 4
        if periph(group) <> 0 then
            poke(pin(1),periph(group))
        else
            print "\nGroup mismatch"
        endif
    endif
endf
